void cleanEeprom(void){
	for(int c=0;c<0x10000;c++) eeprom[c]=0xff;
}

void cleanFlash(void){
	fn_rec *act,*next;
	for(int i=0;i<0x10000;i++){
		flash[i].opcode=0xffff;
		flash[i].flags=0;
		flash[i].src_file=0;
		flash[i].src_line=0;
	}
	act=fn_root;
	while(act){
		next=act->next;
		free(act);
		act=next;
	}
	fn_root=0;
	sourcefile=-1;
	sourceline=-1;
	pc=0;
}

const char *getSrcFileName(int i){
	fn_rec *act;
	act=fn_root;
	while(i){
		act=act->next;
		if(!act) return "Filename not found";
		i--;
	}
	return act->name;
}

/* Test, ob Datei existiert */
int testFile(char *fn){
	FILE *f;
	f=fopen(fn,"r");
	if(!f) return 0;
	fclose(f);
	return 1;
}

/* Ist der Dateiname komplett, oder muss ein Verzeichnis davor? */
void completeFilename(char *fn, const char *root, const char *path){
	const char *cp1;
	char st[PATH_MAX],*cp2;
	if(testFile(fn)) return;
	strcpy(st,root);
	strcat(st,fn);
	if(testFile(st)){
		strcpy(fn,st);
		return;
	}
	/* Dann den Pfad durchtesten */
	cp1=path;
	cp2=st;
	while(*cp1){
		*cp2=*cp1;
		cp1++;
		cp2++;
		if((*cp1==':')||(*cp1==0)){
			if(*cp1) cp1++;
			*cp2=0;
			strcat(st,"/");
			strcat(st,fn);
			if(testFile(st)){
				strcpy(fn,st);
				return;
			}
			cp2=st;
		}
	}
	sprintf(st,"Include File %s not found!",fn);
	ui->statusline->value(st);
	return;
}

void readObj(char *fn){
	fn_rec *prev,*act,*cr;
	FILE *f;
	unsigned char buf[100],*ucp,uch;
	char ch,*cp;
	char root[PATH_MAX],st[PATH_MAX];
	int r,bytectr;
	int fn_start;
	int adr,opc,dat,zn;
	f=fopen(fn,"rb");
	if(!f) return;
	r=fread(buf,0x1a,1,f);
	bytectr=0x1a;
	fn_start=(buf[2]<<8)+buf[3];
	while(bytectr<fn_start){
		r=fread(buf,9,1,f);
		adr=(buf[0]<<16)+(buf[1]<<8)+buf[2];
		opc=(buf[3]<<8)+buf[4];
		dat=buf[5];
		zn=(buf[6]<<8)+buf[7];
		flash[adr].flags=PM_SOURCE;
		flash[adr].opcode=opc;
		flash[adr].src_file=dat;
		flash[adr].src_line=zn;
		bytectr+=9;
	}
	/* Bezugsverzeichnis aus Arbeitsdatei? */
	strcpy(root,fn);
	cp=strrchr(root,'/');
	if(!cp) cp=strrchr(root,'\\');
	if(cp){
		cp++;
		*cp=0;
	}
	else{
		cp=getcwd(root,PATH_MAX);
		strcat(root,"/");
	}
	//printf("root:%s\n",root);
	/* versuchen, die Dateinamen einzulesen */
	cp=st;
	while(ch=fgetc(f)){
		do{
			*cp=ch;
			cp++;
			ch=fgetc(f);
		} while(ch);
		*cp=0;
		completeFilename(st,root,includePath);
		//printf("%s\n",st);
		cr=(fn_rec*)calloc(sizeof(fn_rec),1);
		strcpy(cr->name,st);
		prev=0;
		act=fn_root;
		while(act){
			prev=act;
			act=act->next;
		}
		if(prev) prev->next=cr;
		else fn_root=cr;
		cp=st;
	}
	fclose(f);
	strcpy(fn_hexFile,fn);
	strcpy(st,fn);
	cp=strrchr(st,'/');
	if(!cp) cp=strrchr(st,'\\');
	if(!cp) cp=st; else cp++;
	strcpy(fileshortname,cp);
	refreshTitle();
}

void readHex(char *fnam){
	FILE *f;
	char st[1024],*cp,fn[2*PATH_MAX];
	int len,adr,w,b1,b2;
	strcpy(fn,fnam);
	f=fopen(fn,"r");
	if(!f) return;
	while(fgets(st,1010,f)){
		//printf("%s",st);
		cp=st+1;
		sscanf(cp,"%2x",&len);
		cp+=2;
		sscanf(cp,"%4x",&adr);
		cp+=6;
		adr>>=1;
		//printf("adr=%04x len=%02x\n",adr,len);
		while(len>0){
			sscanf(cp,"%02x%02x",&b1,&b2);
			w=b1|b2<<8;
			cp+=4;
			//printf("%04X: %04X\n",adr,w);
			flash[adr].opcode=w;
			len-=2;
			adr++;
			if(max<adr) max=adr;
		}
	}
	fclose(f);
	strcpy(fn_hexFile,fn);
	cp=strrchr(fn,'/');
	if(!cp) cp=strrchr(fn,'\\');
	if(!cp) cp=fn; else cp++;
	strcpy(fileshortname,cp);
	refreshTitle();
}

void readCodeFile(char *fn){
	char ext[PATH_MAX],*cp;
	cp=strrchr(fn,'.');
	if(!cp) return;
	strcpy(ext,cp+1);
	cp=ext;
	while(*cp){
		*cp=tolower(*cp);
		cp++;
	}
	cleanFlash();
	if(!strcmp(ext,"hex")){
		readHex(fn);
		return;
	}
	if(!strcmp(ext,"obj")){
		readObj(fn);
		return;
	}
	printf("unknown extension %s\n",ext);
}

void loadCodeFile(void){
	char *cp;
	cp=fl_file_chooser("Select Code File",
		"*.{hex,obj}",
		".",
		0);
	if(!cp) return;
	readCodeFile(cp);
	refresh();
}

void createStarter(void)
{
#ifndef WIN32
	int r;
	char starterPath[PATH_MAX];
	char appPath[PATH_MAX],*cp;
	char dest[PATH_MAX];
	FILE *f;
	pid_t pid = getpid();
	cp=getenv("HOME");
	if(!cp) return;
	strcpy(starterPath,cp);
	strcat(starterPath,"/.local/share/applications/avrsim.desktop");
	r=fl_choice("Create Starter  %s?","Yes","No","?",starterPath);
	if(r==1) return;
	if(r==2){
		return;
	}
	f=fopen(starterPath,"w");
	if(!f) return;
	fprintf(f,"[Desktop Entry]\nVersion=1.0\nEncoding=UTF-8\nName=avr-simulator\nGenericName=avr-simulator\nComment=avr-simulator\n");
	sprintf(appPath, "/proc/%d/exe", pid);
	if (readlink(appPath, dest, PATH_MAX) == -1) return;
	fprintf(f,"Exec=%s %%f\n",dest);
	if(cp=strrchr(dest,'/')) strcpy(cp,"/images/avrsim.xpm");
	fprintf(f,"Icon=%s\n",dest);
	fprintf(f,"Terminal=false\nType=Application\nStartupNotify=false\nName[de]=avr-simulator\nGenericName[de]=avr-simulator\n");
	fprintf(f,"Comment[de]=avr-simulator\nName[de_DE]=avr-simulator\nMimeType=application/obj\n");
	fclose(f);
	r=chmod(starterPath,S_IRUSR|S_IWUSR|S_IXUSR|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH);
	if(r) fl_alert("chmod error");
#endif
}
